{--------------------------------------------------------}
{                                                        }
{         The TBezier component is Copyright  1996      }
{         by Cristian Jansenson All Rights Reserved      }
{                                                        }
{--------------------------------------------------------}

unit Bezier;

interface

Uses
  Controls, Graphics, Classes, Messages, Forms, WinProcs;
Type
TStyleBezier = (sbEdit, sbNoEdit);
TBezierChangeEvent = procedure(Sender:TObject; IndexPoint: Integer;
                               ControlPoint: Boolean) of object;

TBezier = class(TGraphicControl)
private
  FBorder: Boolean;
  FColor: TColor;
  FControlX1: Integer;
  FControlY1: Integer;
  FControlX2: Integer;
  FControlY2: Integer;
  FStretch: Boolean;
  FStyle: TStyleBezier;
  FX1: Integer;
  FY1: Integer;
  FX2: Integer;
  FY2: Integer;
  FOnChange: TBezierChangeEvent;

  CanMoveX1Y1: Boolean;
  CanMoveControlX1Y1: Boolean;
  CanMoveControlX2Y2: Boolean;
  CanMoveX2Y2: Boolean;

  FBrush: TBrush;
  FPen: TPen;
  FirstTime: Boolean;

  Procedure DrawCurve;
  Procedure PreventMove;
  procedure RepaintRequest(Sender: TObject);
  Procedure SetBorder(Value: Boolean);
  procedure Setbounds(ALeft, ATop, AWidth, AHeight: Integer); Override;
  procedure SetColor(Value: TColor);
  Procedure SetControlX1(Value: Integer);
  Procedure SetControlX2(Value: Integer);
  Procedure SetControlY1(Value: Integer);
  Procedure SetControlY2(Value: Integer);
  Procedure SetInitialPosition;
  procedure SetPen (Value: TPen);
  Procedure SetStretch(Value: Boolean);
  Procedure SetStyle(Value: TStyleBezier);
  Procedure SetX1(Value: Integer);
  Procedure SetX2(Value: Integer);
  Procedure SetY1(Value: Integer);
  Procedure SetY2(Value: Integer);
  procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
  procedure WndProc(var Message: TMessage); Override;
protected
  Procedure paint; Override;
  Procedure BezierChanged(IndexPoint: Integer; ControlPoint: Boolean); dynamic;
public
  Constructor Create(AOwner: TComponent); Override;
  Destructor Destroy; Override;
published
  property Border: Boolean
    read FBorder write SetBorder default True;
  property Color: TColor
    read FColor write SetColor default clWhite;
  property ControlX1: Integer
    read FControlX1 write SetControlX1 Default 0;
  property ControlX2: Integer
    read FControlX2 write SetControlX2;
  property ControlY1: Integer
    read FControlY1 write SetControlY1;
  property ControlY2: Integer
    read FControlY2 write SetControlY2;
  property Pen: TPen
    read FPen write SetPen;
  property Stretch: Boolean
    read FStretch write SetStretch default False;
  property Style: TStyleBezier
    read FStyle write SetStyle default sbEdit;
  property X1: Integer
    read FX1 write SetX1;
  property X2: Integer
    read FX2 write SetX2;
  property Y1: Integer
    read FY1 write SetY1;
  property Y2: Integer
    read FY2 write SetY2;
  property Align;
  property Cursor;
  property OnClick;
  property OnDblClick;
  property OnDragDrop;
  property OnDragOver;
  property OnMouseDown;
  property OnMouseMove;
  property OnMouseUp;
  property OnChange: TBezierChangeEvent
    read FOnChange write FOnChange;
End;

Procedure Register;

implementation

Procedure TBezier.BezierChanged(IndexPoint: Integer; ControlPoint: Boolean);
Begin
  If Assigned( FOnChange ) then
    FOnChange( Self, IndexPoint, ControlPoint);
End;

Constructor TBezier.Create(AOwner: TComponent);
Begin
  {Call Inherited}
  Inherited Create (Aowner);
  FBrush := TBrush.Create;
  FPen := TPen.Create;
  FBrush.OnChange := RepaintRequest;
  FBrush.Style := bsSolid;
  FBrush.Color := clWhite;
  FColor:= clWhite;
  FPen.OnChange := RepaintRequest;
  FPen.Mode := pmCopy;
  FPen.Color := clBlack;
  FStretch:= False;
  FirstTime:= True;
  FStyle:= sbEdit;
  FBorder:= true;
  FX1:= 0;
  ControlStyle:= ControlStyle + [csOpaque];
End;

Destructor TBezier.Destroy;
Begin
  FBrush.Free;
  FPen.Free;
  {Call Inherited}
  Inherited Destroy;
End;

Procedure TBezier.DrawCurve;
Const num_steps = 100;
Var  t, t_sq, t_cb, incr: double;
     r1, r2, r3, r4: double;
     curve_x, curve_y: Integer;
Begin
  If FirstTime then
  Begin
    {The first time it must set an initial position}
    SetInitialPosition;
    FirstTime:= False;
  End;

  With Canvas do
  Begin
    Pen.Assign(FPen);
    Brush.Assign(FBrush);
    Brush.Color:= FColor;

    {Draw control points if the curve can be edited}
    If (csDesigning in ComponentState) or (Style = sbEdit) then
    Begin
      Pen.Width:= 1;
      MoveTo(X1, Y1);
      Ellipse(FX1-3, FY1-3, FX1+3, FY1+3);
      rectangle(FControlX1-3, FControlY1-3, FControlX1+3, FControlY1+3);
      rectangle(FControlX2-3, FControlY2-3, FControlX2+3, FControlY2+3);
      Ellipse(FX2-3, FY2-3, FX2+3, FY2+3);
      Pen.Width:= FPen.Width;
    End;
  End;

  {Draw the Bezier curve}
  Canvas.MoveTo(X1, Y1);
  incr := 1.0 / num_steps;
  t:=incr;
  While t <= 1.01 do
  begin
    t_sq := t * t;
    t_cb := t * t_sq;

    r1 := (1 - 3*t + 3*t_sq -   t_cb)*X1;
    r2 := (    3*t - 6*t_sq + 3*t_cb)*ControlX1;
    r3 := (          3*t_sq - 3*t_cb)*ControlX2;
    r4 := (                     t_cb)*X2;
    curve_x := Round(r1 + r2 + r3 + r4);

    r1 := (1 - 3*t + 3*t_sq -   t_cb)*Y1;
    r2 := (    3*t - 6*t_sq + 3*t_cb)*ControlY1;
    r3 := (          3*t_sq - 3*t_cb)*ControlY2;
    r4 := (                     t_cb)*Y2;
    curve_y := Round(r1 + r2 + r3 + r4);

    Canvas.LineTo(curve_x, curve_y);
    t := t + incr;
  End;

  {Draw red dotted lines if the curve can be edited}
  if (csDesigning in ComponentState) or (Style = sbEdit) then
  With Canvas Do
  Begin
    Pen.Mode:= pmCopy;
    Pen.Style:= psDot;
    Pen.Color:= clRed;
    Pen.Width:= 1;
    MoveTo(X1,Y1);
    LineTo(ControlX1,ControlY1);
    MoveTo(ControlX2,ControlY2);
    LineTo(X2,Y2);
  End;
End;

Procedure TBezier.Paint;
Begin
  With Canvas do
  Begin
    Pen.Assign(FPen);
    Brush.Assign(FBrush);

    {If with border, draw the border}
    If FBorder then
    Begin
      Pen.Width:= 1;
      Pen.Style:= psSolid;
      Pen.Color:= clBlack;
      Pen.Mode:= pmCopy
    End
    Else
      Pen.Style:= psClear;
  End;

  Canvas.Brush.Color:= FColor;
  Canvas.rectangle(0, 0, Width, Height);
  DrawCurve;
End;

{Disable moving the points}
Procedure TBezier.PreventMove;
Begin
  CanMoveX1Y1:= False;
  CanMoveControlX1Y1:= False;
  CanMoveControlX2Y2:= False;
  CanMoveX2Y2:= False;
End;

procedure TBezier.RepaintRequest(Sender: TObject);
Begin
  Repaint;
End;

Procedure TBezier.SetBorder(Value: Boolean);
Begin
  FBorder:= Value;
  RepaintRequest(Self);
End;

procedure TBezier.Setbounds(ALeft, ATop, AWidth, AHeight: Integer);
Begin
  {If stretch = true, resize the curve acordingly}
  If FStretch then
  Begin
    FX1:=Round(FX1*(AWidth/Width));
    FY1:= Round(FY1*(AHeight/Height));
    FControlX1:= Round(FControlX1*(AWidth/Width));
    FControlY1:= Round(FControlY1*(AHeight/Height));
    FControlX2:= Round(FControlX2*(AWidth/Width));
    FControlY2:= Round(FControlY2*(AHeight/Height));
    FX2:= Round(FX2*(AWidth/Width));
    FY2:= Round(FY2*(AHeight/Height));
  End;
  {call default function}
  Inherited Setbounds(ALeft, ATop, AWidth, AHeight);
End;

procedure TBezier.SetColor(Value: TColor);
Begin
  FColor:= Value;
  Invalidate;
End;

procedure TBezier.SetControlX1(Value: Integer);
Begin
  FControlX1:=Value;
  Invalidate;
  BezierChanged(1, True);
End;

procedure TBezier.SetControlX2(Value: Integer);
Begin
  FControlX2:=Value;
  Invalidate;
  BezierChanged(2, True);
End;

procedure TBezier.SetControlY1(Value: Integer);
Begin
  FControlY1:=Value;
  Invalidate;
  BezierChanged(1, True);
End;

procedure TBezier.SetControlY2(Value: Integer);
Begin
  FControlY2:=Value;
  Invalidate;
  BezierChanged(2, True);
End;

Procedure TBezier.SetInitialPosition;
Begin
  IF FX1 = 0 then
  Begin
    FX1:= Round(Width*2/6);
    FY1:= 3*Height div 4;
    FControlX1:= Round(Width*1/6);
    FControlY1:= Height div 4;
    FControlX2:= Round(Width*5/6);
    FControlY2:= 3*Height div 4;
    FX2:= Round(Width*4/6);
    FY2:= Height div 4;
  End;
End;

procedure TBezier.SetPen (Value: TPen);
Begin
  FPen.Assign(Value);
  Invalidate;
End;

Procedure TBezier.SetStretch(Value: Boolean);
Begin
  FStretch:= Value;
  RepaintRequest(Self);
End;

procedure TBezier.SetStyle(Value: TStyleBezier);
Begin
  FStyle:= Value;
  RepaintRequest(Self);
End;

procedure TBezier.SetX1(Value: Integer);
Begin
  FX1:=Value;
  Invalidate;
  BezierChanged(1, False);
End;

procedure TBezier.SetX2(Value: Integer);
Begin
  FX2:=Value;
  Invalidate;
  BezierChanged(2, False);
End;

procedure TBezier.SetY1(Value: Integer);
Begin
  FY1:=Value;
  Invalidate;
  BezierChanged(1, False);
End;

procedure TBezier.SetY2(Value: Integer);
Begin
  FY2:=Value;
  Invalidate;
  BezierChanged(2, False);
End;

{Override WndProc to be able to intercept messages at design time}
procedure TBezier.WndProc(var Message: TMessage);

{Check that the mouse is inside the Control}
Function MouseInsideBounds(XPosition, YPosition: LongInt): Boolean;
Var Limit: Integer;
Begin
  MouseInsideBounds:= false;
  If (Xposition<>0) and (Yposition<>0) then
  Begin
    Limit:=5;
    If PtInRect(Rect(Limit, Limit, Width - Limit, Height - Limit), Point(XPosition, YPosition)) then
      MouseInsideBounds:= true
     Else
      PreventMove;
  End;
End;

Function CanMove: Boolean;
Begin
  CanMove:= False;
  if (CanMoveX1Y1 or CanMoveX2Y2) or (CanMoveControlX2Y2 or CanMoveControlX1Y1) then
    CanMove:= true;
End;

Var MyWMLMouseMove: TWMMouse;
    MessageProcessed:Boolean;
    Form: TForm;
Begin
  MessageProcessed:= False;
  {If the curve can be edited or at design time...}
  If (Style = sbEdit) or (csDesigning in ComponentState) then
  Begin
    {on mouse up...}
    If (Message.Msg = WM_LBUTTONUP) then
    If CanMove then
    Begin
      {Disable moving the points}
      PreventMove;
      {Make sure the focus stays on the control}
      Perform(WM_SETFOCUS, 0, 0);
      {Update the Object Inspector}
      Form:= GetParentForm(Self);
      if (Form <> nil) and (Form.Designer <> nil) then Form.Designer.Modified;
      MessageProcessed:= true;
    End;

    {On Mouse Leave...}
    If (Message.Msg = cm_mouseLeave) then
    Begin
      {Disable moving the points}
      PreventMove;
      MessageProcessed:= true;
    End
    Else
    If MouseInsideBounds(Message.lParamLo, Message.lParamHi) then
    If (Message.Msg = WM_MOUSEMOVE) then
    if CanMove then
    Begin
      MyWMLMouseMove.Msg:= WM_LBUTTONDOWN;
      MyWMLMouseMove.Keys:= 1;
      MyWMLMouseMove.XPos:= Message.lParamLo;
      MyWMLMouseMove.YPos:= Message.lParamHi;
      WMMouseMove(MyWMLMouseMove);
    End;
    End;

    {On left mousedown, check if the mouse is on one of the points}
    If Message.Msg = WM_LBUTTONDOWN  then
    Begin
      {If the mouse is on one of the points, enable moving}
      If (Abs(Message.lParamLo - FX1) <= 3) and (Abs(Message.lParamHi - FY1) <= 3) then
        CanMoveX1Y1:= True
      Else If (Abs(Message.lParamLo - FControlX1) <= 3) and (Abs(Message.lParamHi - FControlY1) <= 3) then
        CanMoveControlX1Y1:= True
      Else If (Abs(Message.lParamLo - FControlX2) <= 3) and (Abs(Message.lParamHi - FControlY2) <= 3) then
        CanMoveControlX2Y2:= True
      Else If (Abs(Message.lParamLo - FX2) <= 3) and (Abs(Message.lParamHi - FY2) <= 3) then
        CanMoveX2Y2:= True;
      If CanMove then MessageProcessed:= true;
  End;

  {Process the default}
  If not MessageProcessed or (not (csDesigning in ComponentState)) then
    inherited WndProc(Message);
End;

{Process MouseMove}
procedure TBezier.WMMouseMove(var Message: TWMMouseMove);
Var FControlState: TControlState;
begin
  if (csDesigning in ComponentState) or (Style = sbEdit) then
  if CanMoveX1Y1 or CanMoveControlX1Y1 or CanMoveControlX2Y2 or CanMoveX2Y2 then
  Begin
    If CanMoveX1Y1 then
    Begin
      X1:= Message.XPos;
      Y1:= Message.YPos;
    End;
    If CanMoveControlX1Y1 then
    Begin
      ControlX1:= Message.XPos;
      ControlY1:= Message.YPos;
    End;

    If CanMoveControlX2Y2 then
    Begin
      ControlX2:= Message.XPos;
      ControlY2:= Message.YPos;
    End;

    If CanMoveX2Y2 then
    Begin
      X2:= Message.XPos;
      Y2:= Message.YPos;
    End;
  End;

end;

Procedure Register;
Begin
  RegisterComponents('Samples', [TBezier]);
End;

end.

